@@ -1,15 +1,16 @@
 # ----------------------------------------------------------------------------
 # pyglet
 # Copyright (c) 2006-2008 Alex Holkner
+# Copyright (c) 2008-2019 pyglet contributors
 # All rights reserved.
-# 
+#
 # Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions 
+# modification, are permitted provided that the following conditions
 # are met:
 #
 #  * Redistributions of source code must retain the above copyright
 #    notice, this list of conditions and the following disclaimer.
-#  * Redistributions in binary form must reproduce the above copyright 
+#  * Redistributions in binary form must reproduce the above copyright
 #    notice, this list of conditions and the following disclaimer in
 #    the documentation and/or other materials provided with the
 #    distribution.
@@ -32,7 +33,7 @@
 # POSSIBILITY OF SUCH DAMAGE.
 # ----------------------------------------------------------------------------
 
-'''Event dispatch framework.
+"""Event dispatch framework.
 
 All objects that produce events in pyglet implement :py:class:`~pyglet.event.EventDispatcher`,
 providing a consistent interface for registering and manipulating event
@@ -94,8 +95,7 @@ attach handlers to it.  You can push several handlers at once::
 If your function handlers have different names to the events they handle, use
 keyword arguments::
 
-    dispatcher.push_handlers(on_resize=my_resize,
-                             on_key_press=my_key_press)
+    dispatcher.push_handlers(on_resize=my_resize, on_key_press=my_key_press)
 
 After an event handler has processed an event, it is passed on to the
 next-lowest event handler, unless the handler returns `EVENT_HANDLED`, which
@@ -134,34 +134,59 @@ updates the application state and checks for new events::
 Not all event dispatchers require the call to ``dispatch_events``; check with
 the particular class documentation.
 
-'''
+.. note::
+
+    In order to prevent issues with garbage collection, the
+    :py:class:`~pyglet.event.EventDispatcher` class only holds weak
+    references to pushed event handlers. That means the following example
+    will not work, because the pushed object will fall out of scope and be
+    collected::
+
+        dispatcher.push_handlers(MyHandlerClass())
+
+    Instead, you must make sure to keep a reference to the object before pushing
+    it. For example::
+
+        my_handler_instance = MyHandlerClass()
+        dispatcher.push_handlers(my_handler_instance)
+
+"""
 from builtins import object
-from past.builtins import basestring
 
 __docformat__ = 'restructuredtext'
 __version__ = '$Id$'
 
+import sys
 import inspect
+from functools import partial
+import sys
+from .compat import WeakMethod
 
-EVENT_HANDLED = True 
+# PYTHON2 - remove this legacy backwards compatibility hack:
+if sys.version_info < (3, 2):
+    inspect.getfullargspec = inspect.getargspec
+
+EVENT_HANDLED = True
 EVENT_UNHANDLED = None
 
+
 class EventException(Exception):
-    '''An exception raised when an event handler could not be attached.
-    '''
+    """An exception raised when an event handler could not be attached.
+    """
     pass
 
+
 class EventDispatcher(object):
-    '''Generic event dispatcher interface.
+    """Generic event dispatcher interface.
 
     See the module docstring for usage.
-    '''
+    """
     # Placeholder empty stack; real stack is created only if needed
     _event_stack = ()
 
     @classmethod
     def register_event_type(cls, name):
-        '''Register an event type with the dispatcher.
+        """Register an event type with the dispatcher.
 
         Registering event types allows the dispatcher to validate event
         handler names as they are attached, and to search attached objects for
@@ -171,21 +196,21 @@ class EventDispatcher(object):
             `name` : str
                 Name of the event to register.
 
-        '''
+        """
         if not hasattr(cls, 'event_types'):
             cls.event_types = []
         cls.event_types.append(name)
         return name
 
     def push_handlers(self, *args, **kwargs):
-        '''Push a level onto the top of the handler stack, then attach zero or
+        """Push a level onto the top of the handler stack, then attach zero or
         more event handlers.
 
         If keyword arguments are given, they name the event type to attach.
         Otherwise, a callable's `__name__` attribute will be used.  Any other
         object may also be specified, in which case it will be searched for
         callables with event names.
-        '''
+        """
         # Create event stack if necessary
         if type(self._event_stack) is tuple:
             self._event_stack = []
@@ -195,33 +220,41 @@ class EventDispatcher(object):
         self.set_handlers(*args, **kwargs)
 
     def _get_handlers(self, args, kwargs):
-        '''Implement handler matching on arguments for set_handlers and
+        """Implement handler matching on arguments for set_handlers and
         remove_handlers.
-        '''
-        for object in args:
-            if inspect.isroutine(object):
+        """
+        for obj in args:
+            if inspect.isroutine(obj):
                 # Single magically named function
-                name = object.__name__
+                name = obj.__name__
                 if name not in self.event_types:
                     raise EventException('Unknown event "%s"' % name)
-                yield name, object
+                if inspect.ismethod(obj):
+                    yield name, WeakMethod(obj, partial(self._remove_handler, name))
+                else:
+                    yield name, obj
             else:
                 # Single instance with magically named methods
-                for name in dir(object):
+                for name in dir(obj):
                     if name in self.event_types:
-                        yield name, getattr(object, name)
+                        meth = getattr(obj, name)
+                        yield name, WeakMethod(meth, partial(self._remove_handler, name))
+
         for name, handler in kwargs.items():
             # Function for handling given event (no magic)
             if name not in self.event_types:
                 raise EventException('Unknown event "%s"' % name)
-            yield name, handler
+            if inspect.ismethod(handler):
+                yield name, WeakMethod(handler, partial(self._remove_handler, name))
+            else:
+                yield name, handler
 
     def set_handlers(self, *args, **kwargs):
-        '''Attach one or more event handlers to the top level of the handler
+        """Attach one or more event handlers to the top level of the handler
         stack.
-        
+
         See :py:meth:`~pyglet.event.EventDispatcher.push_handlers` for the accepted argument types.
-        '''
+        """
         # Create event stack if necessary
         if type(self._event_stack) is tuple:
             self._event_stack = [{}]
@@ -230,7 +263,7 @@ class EventDispatcher(object):
             self.set_handler(name, handler)
 
     def set_handler(self, name, handler):
-        '''Attach a single event handler.
+        """Attach a single event handler.
 
         :Parameters:
             `name` : str
@@ -238,7 +271,7 @@ class EventDispatcher(object):
             `handler` : callable
                 Event handler to attach.
 
-        '''
+        """
         # Create event stack if necessary
         if type(self._event_stack) is tuple:
             self._event_stack = [{}]
@@ -246,24 +279,26 @@ class EventDispatcher(object):
         self._event_stack[0][name] = handler
 
     def pop_handlers(self):
-        '''Pop the top level of event handlers off the stack.
-        '''
+        """Pop the top level of event handlers off the stack.
+        """
         assert self._event_stack and 'No handlers pushed'
 
         del self._event_stack[0]
 
     def remove_handlers(self, *args, **kwargs):
-        '''Remove event handlers from the event stack.
+        """Remove event handlers from the event stack.
 
-        See :py:meth:`~pyglet.event.EventDispatcher.push_handlers` for the accepted argument types.  All handlers
-        are removed from the first stack frame that contains any of the given
-        handlers.  No error is raised if any handler does not appear in that
-        frame, or if no stack frame contains any of the given handlers.
+        See :py:meth:`~pyglet.event.EventDispatcher.push_handlers` for the
+        accepted argument types. All handlers are removed from the first stack
+        frame that contains any of the given handlers. No error is raised if
+        any handler does not appear in that frame, or if no stack frame
+        contains any of the given handlers.
 
-        If the stack frame is empty after removing the handlers, it is 
+        If the stack frame is empty after removing the handlers, it is
         removed from the stack.  Note that this interferes with the expected
-        symmetry of :py:meth:`~pyglet.event.EventDispatcher.push_handlers` and :py:meth:`~pyglet.event.EventDispatcher.pop_handlers`.
-        '''
+        symmetry of :py:meth:`~pyglet.event.EventDispatcher.push_handlers` and
+        :py:meth:`~pyglet.event.EventDispatcher.pop_handlers`.
+        """
         handlers = list(self._get_handlers(args, kwargs))
 
         # Find the first stack frame containing any of the handlers
@@ -294,11 +329,12 @@ class EventDispatcher(object):
             self._event_stack.remove(frame)
 
     def remove_handler(self, name, handler):
-        '''Remove a single event handler.
+        """Remove a single event handler.
 
         The given event handler is removed from the first handler stack frame
         it appears in.  The handler must be the exact same callable as passed
-        to `set_handler`, `set_handlers` or :py:meth:`~pyglet.event.EventDispatcher.push_handlers`; and the name
+        to `set_handler`, `set_handlers` or
+        :py:meth:`~pyglet.event.EventDispatcher.push_handlers`; and the name
         must match the event type it is bound to.
 
         No error is raised if the event handler is not set.
@@ -308,7 +344,7 @@ class EventDispatcher(object):
                 Name of the event type to remove.
             `handler` : callable
                 Event handler to remove.
-        '''
+        """
         for frame in self._event_stack:
             try:
                 if frame[name] == handler:
@@ -317,9 +353,22 @@ class EventDispatcher(object):
             except KeyError:
                 pass
 
+    def _remove_handler(self, name, handler):
+        """Used internally to remove all handler instances for the given event name.
+
+        This is normally called from a dead ``WeakMethod`` to remove itself from the
+        event stack.
+        """
+        # Iterate over a copy as we might mutate the list
+        for frame in list(self._event_stack):
+            if name in frame and frame[name] == handler:
+                del frame[name]
+                if not frame:
+                    self._event_stack.remove(frame)
+
     def dispatch_event(self, event_type, *args):
-        '''Dispatch a single event to the attached handlers.
-        
+        """Dispatch a single event to the attached handlers.
+
         The event is propagated to all handlers from from the top of the stack
         until one returns `EVENT_HANDLED`.  This method should be used only by
         :py:class:`~pyglet.event.EventDispatcher` implementors; applications should call
@@ -337,49 +386,55 @@ class EventDispatcher(object):
                 Arguments to pass to the event handler.
 
         :rtype: bool or None
-        :return: (Since pyglet 1.2) `EVENT_HANDLED` if an event handler 
+        :return: (Since pyglet 1.2) `EVENT_HANDLED` if an event handler
             returned `EVENT_HANDLED`; `EVENT_UNHANDLED` if one or more event
             handlers were invoked but returned only `EVENT_UNHANDLED`;
-            otherwise ``False``.  In pyglet 1.1 and earler, the return value
+            otherwise ``False``.  In pyglet 1.1 and earlier, the return value
             is always ``None``.
 
-        '''
-        assert event_type in self.event_types, "%r not found in %r.event_types == %r" % (event_type, self, self.event_types)
+        """
+        assert hasattr(self, 'event_types'), (
+            "No events registered on this EventDispatcher. "
+            "You need to register events with the class method "
+            "EventDispatcher.register_event_type('event_name')."
+        )
+        assert event_type in self.event_types,\
+            "%r not found in %r.event_types == %r" % (event_type, self, self.event_types)
 
         invoked = False
 
         # Search handler stack for matching event handlers
         for frame in list(self._event_stack):
             handler = frame.get(event_type, None)
-            if handler:
-                try:
-                    invoked = True
-                    if handler(*args):
-                        return EVENT_HANDLED
-                except TypeError:
-                    self._raise_dispatch_exception(event_type, args, handler)
-
-
-        # Check instance for an event handler
-        if hasattr(self, event_type):
+            if not handler:
+                continue
+            if isinstance(handler, WeakMethod):
+                handler = handler()
+                assert handler is not None
             try:
                 invoked = True
-                if getattr(self, event_type)(*args):
+                if handler(*args):
                     return EVENT_HANDLED
-            except TypeError:
-                self._raise_dispatch_exception(
-                    event_type, args, getattr(self, event_type))
+            except TypeError as exception:
+                self._raise_dispatch_exception(event_type, args, handler, exception)
 
-        if invoked:
+        # Check instance for an event handler
+        try:
+            func = getattr(self, event_type, None)
+            if func:
+                func(*args)
+                return EVENT_HANDLED
+        except BaseException as exception:
+            self._raise_dispatch_exception(event_type, args, getattr(self, event_type), exception)
             return EVENT_UNHANDLED
 
         return False
 
-    def _raise_dispatch_exception(self, event_type, args, handler):
+    def _raise_dispatch_exception(self, event_type, args, handler, exception):
         # A common problem in applications is having the wrong number of
         # arguments in an event handler.  This is caught as a TypeError in
         # dispatch_event but the error message is obfuscated.
-        # 
+        #
         # Here we check if there is indeed a mismatch in argument count,
         # and construct a more useful exception message if so.  If this method
         # doesn't find a problem with the number of arguments, the error
@@ -388,8 +443,11 @@ class EventDispatcher(object):
         n_args = len(args)
 
         # Inspect the handler
-        handler_args, handler_varargs, _, handler_defaults = \
-            inspect.getargspec(handler)
+        argspecs = inspect.getfullargspec(handler)
+        handler_args = argspecs.args
+        handler_varargs = argspecs.varargs
+        handler_defaults = argspecs.defaults
+
         n_handler_args = len(handler_args)
 
         # Remove "self" arg from handler if it's a bound method
@@ -401,30 +459,27 @@ class EventDispatcher(object):
             n_handler_args = max(n_handler_args, n_args)
 
         # Allow default values to overspecify arguments
-        if (n_handler_args > n_args and 
-            handler_defaults and
+        if (n_handler_args > n_args and handler_defaults and
             n_handler_args - len(handler_defaults) <= n_args):
             n_handler_args = n_args
 
         if n_handler_args != n_args:
             if inspect.isfunction(handler) or inspect.ismethod(handler):
-                descr = '%s at %s:%d' % (
-                    handler.__name__,
-                    handler.__code__.co_filename,
-                    handler.__code__.co_firstlineno)
+                descr = "'%s' at %s:%d" % (handler.__name__,
+                                           handler.__code__.co_filename,
+                                           handler.__code__.co_firstlineno)
             else:
                 descr = repr(handler)
-            
-            raise TypeError(
-                '%s event was dispatched with %d arguments, but '
-                'handler %s has an incompatible function signature' % 
-                (event_type, len(args), descr))
+
+            raise TypeError("The '{0}' event was dispatched with {1} arguments, "
+                            "but your handler {2} accepts only {3} arguments.".format(
+                             event_type, len(args), descr, len(handler_args)))
         else:
-            raise
+            raise exception
 
     def event(self, *args):
-        '''Function decorator for an event handler.  
-        
+        """Function decorator for an event handler.
+
         Usage::
 
             win = window.Window()
@@ -439,7 +494,7 @@ class EventDispatcher(object):
             def foo(self, width, height):
                 # ...
 
-        '''
+        """
         if len(args) == 0:                      # @window.event()
             def decorator(func):
                 name = func.__name__
@@ -451,7 +506,7 @@ class EventDispatcher(object):
             name = func.__name__
             self.set_handler(name, func)
             return args[0]
-        elif isinstance(args[0], basestring):   # @window.event('on_resize')
+        elif isinstance(args[0], str):          # @window.event('on_resize')
             name = args[0]
             def decorator(func):
                 self.set_handler(name, func)
