@@ -1,6 +1,7 @@
 # ----------------------------------------------------------------------------
 # pyglet
 # Copyright (c) 2006-2008 Alex Holkner
+# Copyright (c) 2008-2019 pyglet contributors
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -32,8 +33,8 @@
 # POSSIBILITY OF SUCH DAMAGE.
 # ----------------------------------------------------------------------------
 
-'''
-'''
+"""
+"""
 from __future__ import division
 from builtins import chr
 
@@ -45,20 +46,20 @@ import unicodedata
 import warnings
 
 from pyglet import compat_platform
+
 if compat_platform not in ('cygwin', 'win32'):
     raise ImportError('Not a win32 platform.')
 
 import pyglet
-from pyglet.window import BaseWindow, \
-    WindowException, MouseCursor, DefaultMouseCursor, _PlatformEventHandler, \
-    _ViewEventHandler
+from pyglet.window import BaseWindow, WindowException, MouseCursor
+from pyglet.window import DefaultMouseCursor, _PlatformEventHandler, _ViewEventHandler
 from pyglet.event import EventDispatcher
 from pyglet.window import key
 from pyglet.window import mouse
 
 from pyglet.canvas.win32 import Win32Canvas
 
-from pyglet.libs.win32 import _user32, _kernel32, _gdi32
+from pyglet.libs.win32 import _user32, _kernel32, _gdi32, _dwmapi
 from pyglet.libs.win32.constants import *
 from pyglet.libs.win32.winkey import *
 from pyglet.libs.win32.types import *
@@ -81,11 +82,14 @@ _motion_map = {
     (key.DELETE, False):    key.MOTION_DELETE,
 }
 
+
 class Win32MouseCursor(MouseCursor):
     drawable = False
+
     def __init__(self, cursor):
         self.cursor = cursor
 
+
 # This is global state, we have to be careful not to set the same state twice,
 # which will throw off the ShowCursor counter.
 _win32_cursor_visible = True
@@ -129,8 +133,10 @@ class Win32Window(BaseWindow):
                     self._view_event_handlers[message] = func
                 else:
                     self._event_handlers[message] = func
-
-
+                    
+        self._always_dwm = sys.getwindowsversion() >= (6, 2)
+        self._interval = 0
+        
         super(Win32Window, self).__init__(*args, **kwargs)
 
     def _recreate(self, changes):
@@ -160,7 +166,6 @@ class Win32Window(BaseWindow):
         else:
             self._ws_style &= ~(WS_THICKFRAME|WS_MAXIMIZEBOX)
 
-
         if self._fullscreen:
             width = self.screen.width
             height = self.screen.height
@@ -176,7 +181,7 @@ class Win32Window(BaseWindow):
             self._window_class.lpszClassName = u'GenericAppClass%d' % id(self)
             self._window_class.lpfnWndProc = WNDPROC(
                 self._get_window_proc(self._event_handlers))
-            self._window_class.style = CS_VREDRAW | CS_HREDRAW
+            self._window_class.style = CS_VREDRAW | CS_HREDRAW | CS_OWNDC
             self._window_class.hInstance = 0
             self._window_class.hIcon = _user32.LoadIconW(module, MAKEINTRESOURCE(1))
             self._window_class.hbrBackground = black
@@ -225,7 +230,6 @@ class Win32Window(BaseWindow):
                 self._view_window_class.hInstance,
                 0)
 
-
             self._dc = _user32.GetDC(self._view_hwnd)
         else:
             # Window already exists, update it with new style
@@ -293,7 +297,7 @@ class Win32Window(BaseWindow):
 
         _user32.DestroyWindow(self._hwnd)
         _user32.UnregisterClassW(self._window_class.lpszClassName, 0)
-        
+
         self._window_class = None
         self._view_window_class = None
         self._view_event_handlers.clear()
@@ -303,14 +307,30 @@ class Win32Window(BaseWindow):
         self._dc = None
         self._wgl_context = None
         super(Win32Window, self).close()
-
+        
+    def _dwm_composition_enabled(self):
+        """ Checks if Windows DWM is enabled (Windows Vista+)
+            Note: Always on for Windows 8+
+        """
+        is_enabled = c_int()
+        _dwmapi.DwmIsCompositionEnabled(byref(is_enabled))
+        return is_enabled.value
+        
     def _get_vsync(self):
-        return self.context.get_vsync()
+        return bool(self._interval)
     vsync = property(_get_vsync) # overrides BaseWindow property
 
     def set_vsync(self, vsync):
         if pyglet.options['vsync'] is not None:
             vsync = pyglet.options['vsync']
+            
+        self._interval = vsync
+
+        if not self._fullscreen:
+            # Disable interval if composition is enabled to avoid conflict with DWM.
+            if self._always_dwm or self._dwm_composition_enabled():
+                vsync = 0
+
         self.context.set_vsync(vsync)
 
     def switch_to(self):
@@ -318,6 +338,12 @@ class Win32Window(BaseWindow):
 
     def flip(self):
         self.draw_mouse_cursor()
+        
+        if not self._fullscreen:
+            if self._always_dwm or self._dwm_composition_enabled():
+                if self._interval:
+                    _dwmapi.DwmFlush()
+                    
         self.context.flip()
 
     def set_location(self, x, y):
@@ -437,7 +463,7 @@ class Win32Window(BaseWindow):
             raw_mouse.dwFlags = RIDEV_NOLEGACY
             raw_mouse.hwndTarget = self._view_hwnd
         else:
-            raw_mouse.dwFlags |= RIDEV_REMOVE
+            raw_mouse.dwFlags = RIDEV_REMOVE
             raw_mouse.hwndTarget = None
 
         if not _user32.RegisterRawInputDevices(
@@ -447,11 +473,6 @@ class Win32Window(BaseWindow):
 
         self._exclusive_mouse_buttons = 0
         if exclusive and self._has_focus:
-            # Move mouse to the center of the window.
-            self._reset_exclusive_mouse_screen()
-            x, y = self._exclusive_mouse_screen
-            self.set_mouse_position(x, y, absolute=True)
-
             # Clip to client area, to prevent large mouse movements taking
             # it outside the client area.
             rect = RECT()
@@ -459,19 +480,16 @@ class Win32Window(BaseWindow):
             _user32.MapWindowPoints(self._view_hwnd, HWND_DESKTOP,
                                     byref(rect), 2)
             _user32.ClipCursor(byref(rect))
+            # Release mouse capture in case is was acquired during mouse click
+            _user32.ReleaseCapture()
         else:
             # Release clip
             _user32.ClipCursor(None)
-            # Move mouse back to the middle of the client area.
-            if self._exclusive_mouse_screen:
-                x, y = self._exclusive_mouse_screen
-                self.set_mouse_position(x, y, absolute=True)
 
         self._exclusive_mouse = exclusive
         self._exclusive_mouse_focus = self._has_focus
         self.set_mouse_platform_visible(not exclusive)
 
-
     def set_mouse_position(self, x, y, absolute=False):
         if not absolute:
             rect = RECT()
@@ -1024,11 +1042,6 @@ class Win32Window(BaseWindow):
         return 0
     '''
 
-    @Win32EventHandler(WM_MOUSEACTIVATE)
-    def _event_mouse_activate(self, msg, wParam, lParam):
-        # The user clicked on the Window. Activate it but eat the click event.
-        return MA_ACTIVATEANDEAT
-
     @Win32EventHandler(WM_SETFOCUS)
     def _event_setfocus(self, msg, wParam, lParam):
         self.dispatch_event('on_activate')
